from typing import TypeVar, Optional
from xmodel.remote import XRemoteError
from xmodel.remote.model import RemoteModel
from xsentinels import Default

from xdynamo.api import DynApi, lazy_load_types_for_dyn_api
from xdynamo.common_types import DynKey
from xmodel.base.model import Self

from xdynamo.errors import XModelDynamoNoHashKeyDefinedError

M = TypeVar('M')


class DynModel(
    RemoteModel,
    dyn_name=None,
    dyn_service=Default,
    dyn_environment=Default,
    lazy_loader=lazy_load_types_for_dyn_api
):
    """
    Used to easily parse/generate JSON from xyn_sdk model's for use in Dynamo.
    So it will take advantage of all the other features of the xyn_sdk models.
    This includes automatically converting dates to/from strings, converting strings
    to numbers, looking up child-objects from other tables automatically, etc.
    It also will modify the results JSON so remove blank values; to easily prevent
    these sorts of errors in dynamo boto3 library.

    We pass in None for name/service to indicate we don't have an associated table,
    that we are more of an abstract class.
    """
    api: DynApi[Self]
    id: str

    @property
    def id(self) -> Optional[str]:
        # We could do some intelligent caching, but for now just calculate each time.
        try:
            return DynKey.via_obj(self).id
        except XModelDynamoNoHashKeyDefinedError:
            # There is something wrong with class structure, there is no hash-key defined!
            raise
        except XRemoteError:
            # Any other error, we simply don't have a full `id` value assigned to object.
            return None

    @id.setter
    def id(self, value):
        structure = self.api.structure
        if type(value) is str:
            parsed_value = value.split('|')
            hash_value = parsed_value[0]
            range_value = parsed_value[1] if len(parsed_value) == 2 else None

            self.__setattr__(structure.dyn_hash_field.name, hash_value)
            if range_value:
                self.__setattr__(structure.dyn_range_field.name, range_value)
            return

        raise NotImplementedError(
            "Read-only for now, but want to support it. "
            "Supporting it would involve parsing ID with DynKey, and taking hash/range key "
            "components and setting them on the proper attributes."
            "\n\n"
            "Also, want to eventually support for using 'id' as a HashField "
            "(ie: a single/only key called 'id' in dynamo-db)"
        )



"""
Used to keep track of a shared connection and Dynamo table resources.

If you want to easily send/get information from Dynamo via orm.

Important Classes

- `DynamoDB`: A resource (see `xyn_resource` for more details) that keeps
    track of a shared boto3 dynamo session/connection. It also caches the fact if a table
    is created (ie: no need to check every time, just the first time we use a table).

- `DynamoTableCreator`: Interface used by `DynamoDB` to have an easy way to initiate the creation
    of a Dynamo table if needed (if it does not exist).

"""

from __future__ import annotations

import logging
from typing import Any, Dict, Optional

from botocore.exceptions import ClientError
from xinject import Dependency

from xboto.resource import dynamodb
from .errors import XModelDynamoError

log = logging.getLogger(__name__)

__all__ = ["DynamoTableCreator", "DynamoDB"]

_auto_create_table_only_in_environments = {'unittest', 'local'}


class DynamoTableCreator:
    """A callable that can, when called, create the needed Dynamo table, and then return the newly
    created table as a boto3 table resource [via: return dynamo.db.create_table(...)].
    """

    def __call__(self, dynamo: DynamoDB) -> Any:
        """Something like this::

        return dynamo.db.create_table(TableName="SomeName", KeySchema=[...], ...)
        """
        raise NotImplementedError("Must Implement To Create Dynamo Table.")


class DynamoDB(Dependency, attributes_to_skip_while_copying=["_table", "_verified", "_db"]):
    """
    Resource that represents a DynamoDB connection.  This allows us to pool the dynamo
    connection among everything that uses Dynamo, so that we can reuse existing connections.
    This speeds up Dynamo by a fair amount.  Every time you open a connection with Dynamo,
    it has to figure out how to get the encryption key, but subsequent requests using the same
    connection don't have to.
    """
    _tables: Dict[str, Any]
    _verified: Dict[str, bool]

    def __init__(self):
        self._tables = {}
        self._verified = {}

    @property
    def db(self):
        return dynamodb

    def table(self, name: str, table_creator: Optional[DynamoTableCreator] = None):
        """Returns existing table or creates + returns one if we don't have the resource
        currently. When we create a new one, we will remember it in a weak-fashion in
        case we get asked again. Once all other references to the table resource are
        gone the python garbage-collector will cleanup our reference automatically.
        If that happens and we get asked for that table_resource, we will create and
        return a new one [and remember if for future use].

        The reason to provide a table_creator is so you can easily Mock DynamoDB, as it
        will be created in the 'Mock' framework automatically.

        Also when devs run the code locally without deploying it, it will create the table
        for them.

        When you deploy code via serverless, it should be the thing the creates the table
        for the deployed environment [ie: testing/prod/etc].  The table should already exist
        that way when the app runs, and so it does **NOT** need table-creating AWS permissions.

        Args:
            name: Name of the table to get a resource for.
            table_creator (Union[DynamoTableCreator, None]): ...
                If None: Won't verify table is created or ready.

                If DynamoTableCreator:
                    Verifies the table exists and is ready. If it does not exist we create
                    the table by 'calling' whats passed in here and wait for it to be ready.
                    If the table is in a status that indicates it can't be used at the moment
                    [example: If table is 'DELETING'], we raise an XynLibError.
        """
        verified = self._verified.get(name, False)

        try:
            from xcon import xcon_settings
            # We only verify/create-table-if-needed in specific environments.
            if xcon_settings.environment not in _auto_create_table_only_in_environments:
                verified = True
        except ImportError:
            # If `xcon` unavailable, just assume we don't want to auto-create tables
            # todo: Put in a configurable setting that allows one to turn on/off
            #       auto-table-creation.
            #       (and some way to communicate billing mode???).
            #
            # todo: Log about why not creating tables, but log it only once.
            verified = True

        table = self._tables.get(name)
        if table is not None and verified:
            return table

        if table is None:
            table = dynamodb.Table(name)

        if verified:
            self._tables[name] = table
            return table

        if not table_creator:
            # Don't verify table if we don't have a table creator, just return it.
            self._tables[name] = table
            self._verified[name] = False
            return table

        try:
            log.info(f"Getting Table Status for ({name}).")
            status = table.table_status
            # It turns out, a table is still usable while "UPDATING", so only worry about
            # DELETING and CREATING.
            if status == "CREATING":
                log.warning(
                    f"Dynamo status for table ({name}) is CREATING; based on past experience "
                    f"we can still use the table [to at least read values] if the table is "
                    f"being restored from backup while it's in the CREATING status. So I am not "
                    f"going to wait for table to become ACTIVE, I'll try to use the table "
                    f"immediately."
                )
                # This is how we could wait for the table, disabling for now, see log.warning ^^^
                # table.wait_until_exists()
            elif status == "DELETING":
                raise XModelDynamoError(f"Dynamo Table ({name}) status is 'DELETING'???")
        except ClientError:
            # This means the table has not been created yet, we create and wait for it to exist.
            log.warning(f"Dynamo table ({name}) does not exist, creating...")
            table = table_creator(dynamo=self)
            table.wait_until_exists()

        self._tables[name] = table
        self._verified[name] = True
        return table



from typing import Dict, Optional, Any, Set, TYPE_CHECKING
from boto3.dynamodb.table import BatchWriter

from xmodel.remote import XRemoteError
from xinject import DependencyPerThread

if TYPE_CHECKING:
    from xdynamo.api import DynApi


class _DynBatcher(object):
    """
    Used by ``Table`` as the context manager for batch writes.

    You likely don't want to try to use this object directly.
    """
    _table_to_boto_writer: Dict[Any, BatchWriter]
    _enter_count = 0
    _dyn_batch_resources_added_to: Set['_DynBatchResource']

    def __init__(self):
        self._dyn_batch_resources_added_to = set()
        self._table_to_boto_writer = dict()

    def batch_writer(self, api: 'DynApi') -> BatchWriter:
        if self._enter_count <= 0:
            raise XRemoteError("Must use DynBatch via `with` statement as a context manager.")

        table = api.table

        writer_map = self._table_to_boto_writer
        table_id = id(table)
        batch_writer = writer_map.get(table_id)
        if batch_writer:
            return batch_writer

        structure = api.structure

        pkeys = [structure.dyn_hash_key_name]
        range_key = structure.dyn_range_key_name
        if range_key:
            pkeys.append(range_key)

        batch_writer = table.batch_writer(overwrite_by_pkeys=pkeys)
        writer_map[table_id] = batch_writer
        # Activate writer, we are
        batch_writer.__enter__()
        return batch_writer

    def __enter__(self):
        enter_count = self._enter_count
        if enter_count <= 0 and self._table_to_boto_writer:
            # If we have writers at this point, there is a serious problem.
            # We may have left over writers
            raise XRemoteError(
                "We got 'entered' as a context manager, but we have writers. Having writers at "
                "this point is a serious problem. We may have left-over writers from a previous "
                "use as a context manager [via `with`]?"
            )

        if enter_count < 0:
            raise XRemoteError(
                "We got 'entered' as a context manager, but we have a negative enter count of "
                f"({enter_count})! This indicates a problem in DynBatch."
            )

        try:
            if enter_count > 0:
                return

            resource = _DynBatchResource.grab()
            resource.add_writer(self)
            self._dyn_batch_resources_added_to.add(resource)
        finally:
            self._enter_count += 1

    def __exit__(self, type, value, traceback):
        enter_count = self._enter_count
        enter_count -= 1
        if enter_count < 0:
            raise XRemoteError("DynBatch enter/exit count is below zero, we got unbalanced.")
        self._enter_count = enter_count
        if enter_count == 0:
            # Exit all batch writers, clear all writers.
            # It's assumed that a writer we have already had '__enter__' called on it.
            for writer in self._table_to_boto_writer.values():
                writer.__exit__(type, value, traceback)

            for resource in self._dyn_batch_resources_added_to:
                resource.remove_writer(self)

            # Remove all internal references, we have cleaned up and closed them all up.
            self._table_to_boto_writer.clear()
            self._dyn_batch_resources_added_to.clear()


class _DynBatchResource(DependencyPerThread):
    # Instead of inheriting from `ThreadUnsafeResource`, we set flag directly ourselves.
    # This allows us to be compatible with both v2 and v3 of xinject.
    #
    # It only makes sense to have _DynBatchResource on a single-thread.
    writers: Dict[id, 'DynBatch']

    def __init__(self):
        self.writers = {}

    def add_writer(self, writer: 'DynBatch'):
        writer_id = id(writer)
        writers = self.writers
        if writer_id in writers:
            raise XRemoteError(
                f"Adding writer {writer} to DynBatchResource, but that writer is "
                "already there!  The writer should not do this, it's state is bad."
            )

        writers[writer_id] = writer

    def remove_writer(self, writer: 'DynBatch'):
        self.writers.pop(id(writer))

    def have_writer(self) -> bool:
        return bool(self.writers)

    def current_writer(self, create_if_none: bool = False) -> Optional['DynBatch']:
        writers = self.writers
        if not writers:
            return DynBatch() if create_if_none else None

        return list(writers.values())[-1]


class DynBatch(_DynBatcher):
    """
    Allows one to batch bulk updates/deletes (via dynamo put/delete-item) with a context manager.
    You can bulk-delete/update currently via:

    - `DynClient.delete_objs`
    - `DynClient.update_objs`

    But if you want to combine a number of separate update/delete object calls
    (including with other calls to `DynClient.delete_objs` / `DynClient.update_objs`)
    into the same request(s), this class allows you to do that.

    For example code, see [Batch Updating Deleting](#batch-updating-deleting).
    """
    pass


class DynTransaction(_DynBatcher):
    """
    .. todo:: In the future, there will be a class called `DynTransaction` that you can use to
        batch transactions together. It would also allow bulk-partial updating/patching objects
        instead of using 'puts' to replace entire object. And to tie a set of objects together
        that must all be written or rolled back in one go.
    """
    pass



from typing import (
    TYPE_CHECKING, TypeVar, Union, Sequence, Iterable, Optional, List, Dict, Any, Set
)
from boto3.dynamodb import conditions
from boto3.dynamodb.table import BatchWriter, TableResource
from xbool import bool_value

from .db import DynamoDB
from xcon import xcon_settings
from logging import getLogger
from .const import CONDITIONAL_CHECK_FAILED_KEY
from xboto.resource import dynamodb
from xmodel.common.types import FieldNames, JsonDict
from xmodel.remote import XRemoteError
from xmodel.base.fields import Converter
from xmodel.remote.client import RemoteClient
from xdynamo.common_types import (
    DynKey, DynParams, _ProcessedQuery, get_dynamo_type_from_python_type
)
from xdynamo.resources import _DynBatchResource
from xsentinels.default import Default
from xurls.url import UrlStr, Query
from xloop import xloop

if TYPE_CHECKING:
    from xdynamo.model import DynModel
    from xdynamo.api import DynApi
    M = TypeVar('M', bound=DynModel)
else:
    M = TypeVar('M')

log = getLogger(__name__)


class DynClient(RemoteClient[M]):
    """
    Skeleton/Placeholder Class for future work, see story and the other classes in
    this file for more details: https://app.clubhouse.io/xyngular/story/13989
    """

    # This is only here to give IDE's a more concrete-class to use for type/code completion.
    # The type-hint is not otherwise used. That var is valid/gettable on an instance.
    # See `xmodel.remote.client.RemoteClient.api` for more details.
    api: 'DynApi[M]'

    def delete_obj(self, obj: Union[M, DynKey], *, condition: Query = None):
        """
         Attempts to delete passed in object from dynamodb table.

         If you pass in a condition, it will be evaluated on the dynamodb-service side and the
         deletes will only happen if the condition is met.
         This can help prevent race conditions if used correctly.

         If there is a batch-writer currently in use, we will try to use that to batch the deletes.

         Keep in mind that if you pass in a `condition`, we can't use the batch write.
         We will instead send of a single-item delete request with the condition attached
         (bypassing any current batch-writer that may be in use).

         Args:
            obj: Object to delete, it MUST have values for it's primary key(s)
                (no other values are required, the primary-key values are the only ones used).
            condition: Optional, if passed in we will send this as a `ConditionExpression`
                to dynamo.  You can pass in a Query, ie: the same structure that is used
                for `self.query()` and `self.scan()`.

                We will take the Query correctly format it for you into the `ConditionExpression`.

                If the condition is met, the item will be deleted.

                If the condition is NOT met, we will handle it for you by catch the exception,
                logging about it, setting error on response_state of object, and then returning.
                Conditions are generally used to help prevent race-conditions,
                ie: to prevent a deleted on purpose.

                Most of the time the intent is to ignore what the result is, it was either
                deleted or not;
                ie: it will eventually be handled in some way it if it was not deleted.

                If you want to know when a conditional delete fails, you can look at
                deleted objects `obj.api.response_state`.

                It's `had_error` will be `True`, there will be some info in `errors`,
                but also a field error with name `_conditional_check` and code `failed`
                you can easily check for via:

                `....response_state.has_field_error('_conditional_check', 'failed')`

                or you can use `xyn_model_dynamo.const.CONDITIONAL_CHECK_FAILED_KEY` for
                the field key.
         """
        # Reset object error response state, we are try afresh.
        obj.api.response_state.reset()

        # Get the DynKey
        # If we don't get a dyn-key, check for that and raise nicer, higher-level error.
        key = obj if isinstance(obj, DynKey) else DynKey.via_obj(obj)
        params = {
            'Key': key.key_as_dict(),
        }

        # To keep things simple, I am using 'put' which replaces entire item,
        # so get all properties of item regardless if they changed or not.
        # todo: Check for primary key and raise a nicer, higher-level exception in that case.

        if not condition:
            resource = self._table_or_batch_writer()
            resource.delete_item(**params)
            return

        # Add the conditional query to the dynamodb params dict...
        # (can't batch-delete things with conditions, so we do them one at a time instead).
        self._add_conditions_from_query(
            query=condition, params=params, filter_key='ConditionExpression'
        )

        try:
            self.api.table.delete_item(**params)
        except dynamodb.meta.client.exceptions.ConditionalCheckFailedException as e:
            # By default, we ignore conditional check failed, because this means we did not want
            # to delete the item on purpose; ie: it's not an error.
            # conditions are mostly used to prevent race-conditions, hence ignored by default.
            # If someone in the future ends up wanting the exception, we can have an argument
            # or some other way to indicate that we should re-raise the condition check exception.
            # See `self.delete_obj` doc comment (above) for more details.
            log.info(
                f"Didn't delete dynamo obj ({obj}) due to condition not met ({condition}), "
                f"exception from dynamo ({e})."
            )
            state = obj.api.response_state
            state.had_error = True
            state.errors = [
                'The conditional check failed, item was not deleted.',
                {'conditional-check-failed-exception': e}
            ]
            state.add_field_error(field=CONDITIONAL_CHECK_FAILED_KEY, code='failed')

    def delete_objs(self, objs: Sequence[Union[M, DynKey]], condition: Query = None):
        """ Uses a batch-writer to put the items. Much more efficient than doing it one at a time.

            If you pass in a `condition`, the batch-writer can't be used and normal single-deletes
            will automatically be used instead.

            If you only give me one item, directly calls `delete_obj` without a batch-writer.
        """
        if not objs:
            return

        if len(objs) == 1:
            self.delete_obj(obj=objs[0], condition=condition)
            return

        with _DynBatchResource.grab().current_writer(create_if_none=True):
            for i in objs:
                self.delete_obj(obj=i, condition=condition)

    # todo: Someday support an iterable for `objs`
    #  (ie: open it wider then a Sequence, ie: to support generators).
    def send_objs(
            self,
            objs: Sequence[M],
            *,
            condition: Query = None,
            url: UrlStr = None,
            send_limit: int = None
    ):
        """
        Used to send any number of objects to Dynamo in as efficient a manner as possible.

        If possible, uses a batch-writer to put the items.
        It's WAY more efficient than doing it one at a time.

        If a condition is supplied, won't use a batch-writer as only one item can be put
        at a time with a condition.

        In the future at some point, will support transactions and at that point could put
        more items per-request at a time with condition(s).

        Args:
            objs: Objects to send to dynamo.
            url: Not used in Dynamo, ignore
            send_limit: Currently unused, we try to push as much as possible.

            condition: Optional, if passed in we will send this as a `ConditionExpression`
                to dynamo.  You can pass in a Query, ie: the same structure that is used
                for `self.query()` and `self.scan()`.

                We will take the Query correctly format it for you into the `ConditionExpression`.

                If the condition is met, the item will be put into table.

                If the condition is NOT met, we will handle it for you by catch the exception,
                logging about it, setting error on response_state of object, and then returning.
                Conditions are generally used to help prevent race-conditions,
                ie: to prevent a put on purpose; so we don't raise an exception.

                Most of the time the intent is to ignore what the result is, it was either
                put into table or not;
                ie: it will eventually be handled in some way it if it was not put into table.

                If you want to know when a conditional delete fails, you can look at
                deleted objects `obj.api.response_state`.

                It's `had_error` will be `True`, there will be some info in `errors`,
                but also a field error with name `_conditional_check` and code `failed`
                you can easily check for via:

                `....response_state.has_field_error('_conditional_check', 'failed')`

                or you can use `xyn_model_dynamo.const.CONDITIONAL_CHECK_FAILED_KEY` for
                the field key.

        """
        if not objs:
            return

        if len(objs) == 1:
            self._put_item(item=objs[0], condition=condition)
            return

        if condition:
            for i in objs:
                self._put_item(item=i, condition=condition)
            return

        with _DynBatchResource.grab().current_writer(create_if_none=True):
            for i in objs:
                self._put_item(item=i)

    def get(
            self,
            query: Query = None,
            *,
            top: int = None,
            fields: FieldNames = Default,
            allow_scan: bool = False,
            **dynamo_params
    ) -> Iterable[M]:
        """
        This is the standard/abstract interface method that all RestClient's support.

        The idea behind this method is to figure out how to route this query in the most
        efficient manner we can do it in. If you want to guarantee a specific type of query
        you can use `DynClient.query`, etc.

        Generally, this is the best, generally method to use since it can adapt your request
        to the most efficient way to query Dynamo. The `DynApi` will use this method generally
        when it's asked to get something (for example, when using `DynApi.get_via_id`).

        For more info on how you can query, see:

        - [Advanced Queries](#advanced-queries)
            - [Examples](#examples_1)

        Here is the general process this method goes though to determine how to query dynamo
        for a given query:

        1. If provided query is empty.
              - Will paginate though all items in table efficiently; multiple items
                will be returned per-request in parallel.
                You'll get back a generator that gives you an object at time, but behind
                the scenes we are getting a page at a time from dynamo.
        2. We will try to batch it en-mass in parallel if we can.
           We can do this if one of the following is true:
              - Query only contains 'id' key; this is because an 'id' should have full primary key;
                meaning it will contain both hash and range key (if needed).
                You can provide a list of strings or `DynKey` objects.
                   - See `DynModel.id` for more details on how to use `str` with an 'id'.
                   - See `DynKey` for more details on what a composite primary-key as a `str` is
                     like.
              - Table only has a hash-key (no range/sort key) and you provide nothing but
                hash-keys (ie: no other attributes in query)
        3. Next, we see if we can use a Dynamo query via `DynClient.query`.
              - This will use multiple queries if/as needed, but while minimizing the number of
                queries that are needed. Just depends on the query that is provided.
                More then one query is needed if there is more then one hash-key in the query,
                and there are other attributes you are filtering on
                (and so #2 above, batch-get, can't be used)
                Doing multiple queries is still far-faster then doing a scan.
                We return objects via the generator and only execute the second (or more)
                queries
              - In the future, this will automatically querying a Secondary Index if one is
                available.
                (Not Implemented Yet, we will add this when needed)
        4. Next, fallback to using a Global Index if one is available.
              - (Not Implemented Yet, we will do it when we need it)
        5. If that's not possible, in the future we fallback to scan operation.
            - (Not Implemented Yet; This isn't implemented currently and will raise NotImplemented.
              Scan operations can be very slow and so it's not something we really want to do
              normally... I did implement this if getting everything, but with filters/attributes
              to query on I decided to wait until we need it before implementing this currently).

        .. todo:: Implement querying via Secondary/Global indexes; Scanning with query values.
            see above for details.



        Args:
            query: Dict keys are the attribute/key names and values are what
                to filter on for that name.  Operators after double `__` work just like you
                would expect for our xyngular API's... here is an example:

                ```python
                { "some_attr_name__gte": 2 }
                ```

                In this case, we look for `some_attr_name` greater than or equal to integer `2`.

                You MUST provide at least one value for the hash key of the table for right now.
                In the future, we will support doing a table-scan to support queries without
                a hash-key. But for right now it's required.

                For more information with examples see [Advanced Queries](#advanced-queries).

            top: This is supposed to only return this mean records; currently not implemented.
            fields: This is supposed to only retrieve provided field names in returned objects;
                currently not implemented.
            allow_scan: Defaults to False, which means this will raise an exception if a scan
                is required to execute your get.  Set to True to allow a scan if needed
                (it will still do a query, batch-get, etc; if it can, it only does a scan
                if there is no other choice).

                If the query is blank or None will still do a scan regardless of what you pass
                (to return all items in the table).
            **dynamo_params: Extra parameters to include on the Dynamo request(s) that are
                generated. This is 100% optional.
        Returns:

        """

        if not query:
            # If no query.... just get all items via a bulk-scan.
            return self._get_all_items()

        # todo: We want basic logic in here to decide on batch_get vs query
        #       vs [eventually] dyn_scan.
        #       We sort of implement basic logic here to determin
        #       the best way to query dynamo so we have that logic mostly centralized instead
        #       of scattered everywhere. The logic is cetnered around how we will mostly be using
        #       dynamo. The times where we have exceptions, the outside world can just directly
        #       call `dyn_scan/query` them selves and customize the dynamo call more.
        #       `get` is the thing that figured out the best to this to in the general
        #       case.
        #   Future Vision of logic flow [this is not fully implemented or final yet]:
        #       1. Only have id/hash/range keys
        #           A. Have more than one id/hash/range keys.
        #               I. We can use the batch method.
        #           B. Use query
        #       2. Have id/hash + other attributes
        #           A. More than one hash
        #               I. Use several query calls.
        #           B. Only one hash
        #               I. Use single query call.
        #       3. Have no hash key, just other attributes.
        #           A. Must use `scan`.
        #               I. This is just something we will do later, don't need `scan` at moment.
        #

        query = _ProcessedQuery.process_query(query, api=self.api)
        structure = self.api.structure
        have_range_key = bool(structure.dyn_range_key_name)

        # Check to see if we only have key fields (without other filtering criteria);
        # if so we can do a batch-get, which is the fastest way to get a number of specific
        # values.  If we
        if query.contains_only_keys():
            dyn_keys = query.dyn_keys()

            # If we have no range-key, they all dyn-gets support get-batch.
            # it's only the lack of range-key, or a non 'eq' operator for range-key
            # that would disqualify a specific DynKey.
            #
            # todo: Some of these keys could support batch-get, we might consider getting
            #   the ones that do via batch-get, get others via query?

            all_support_batch_get = True
            if have_range_key:
                for dyn_key in dyn_keys:
                    if not dyn_key.range_key:
                        all_support_batch_get = False
                        break

                    if dyn_key.range_operator and dyn_key.range_operator not in ('eq', 'is_in'):
                        all_support_batch_get = False
                        break

            # We have just DynKey's, so we can do a batch get (no other conditions/filters).
            # This will automatically batch a 100 at a time for us via a generator.
            # Dynamo will fetch these in parallel!
            if all_support_batch_get and dyn_keys:
                return self.batch_get(keys=dyn_keys)

        # If we have some sort of key(s) we can use (a hash key with an optional range key).
        if query.dyn_keys():
            # todo: Support `top` and `fields`.
            # todo: Support multiple hash-keys [one query per hash key].
            # todo: unless this table does not have a range key [no hash/range to tie].

            # We have a query that has the hash-key in it, that's good enough to use a query.
            return self.query(query=query)

        if allow_scan:
            return self.scan(query=query)

        # todo: Support 'scans' or always raise error? Scans are very expensive.
        # todo: Support Global + Secondary Indexes
        #  (secondary indexes are partially supported now, since they require the hash-key
        #   so we would currently do a query with a filter, and scan whole hash-key/page).
        raise NotImplementedError(
            "There are no hash-keys or id's in query, and I don't support auto routing to a scan "
            "when `allow_scan` is False, or using a global indexes at the moment. "
            "This is what you need to do without a hash-key/id. "
            "Scan operations are slow, so for being conservative to prevent accidentally doing "
            "one. For now you need to explicitly do them via self.scan or pass in True to the "
            "`allow_scan` parameter of this `get` method; unless your "
            "retrieving all records (ie: blank query). This may change in the future, but for "
            "now a scan operation requires an explicit opt-in."
            "\n"
            "TODO: Support Global Indexes - if there is a global index "
            "then we should use them if query is using the global-index hash-key and other"
            "attributes are in global-index as well."
        )

    def batch_get(
            self, keys: Iterable[DynKey], **params: DynParams
    ) -> Iterable[M]:
        """
        Will fetch keys in the largest batch it can at a time it can from Dynamo;
        Dynamo will fetch each page of values in parallel!

        We split up the keys into 100 increments at a time automatically right now
        For each unique hash key in the set of keys provided, Dynamo will parallel fetch
        the keys (if two keys have the same hash but different range key, dynamo will do
        them sequentially).

        In the future, we may attempt to fetch multiple 100 blocks of keys asynchronously.
        As the returned generator/iterable is gone through to increase the speed.
        We don't do that yet.

        Args:
            keys (Iterable[DynKey]): Keys to fetch.
                Can be of any size, a generator will be returned to minimize memory use.

                .. tip:: If you pass in a `set`, we will be slightly more efficient.
                    We need to ensure the results are uniquified, if you pass a set we can skip
                    doing it.

            **params: An optional set of extra parameters to include in request to Dynamo,
                if so desired.

        Returns: An Iterable/Generator that will efficiently paginate though the results for you.

        """
        structure = self.api.structure
        hash_name = structure.dyn_hash_key_name
        range_name = structure.dyn_range_key_name

        if not keys:
            return []

        def batch_pagination_generator(items):
            if not items:
                return xloop()

            table_name = structure.fully_qualified_table_name()

            # We want to merge our items with anything that could already be there...
            copy_params = {**params}
            req_items_param = copy_params.setdefault('RequestItems', {})
            table_items = req_items_param.setdefault(table_name, {})
            table_keys = table_items.setdefault('Keys', [])
            table_keys.extend(items)

            # batch_get_item is only available on dynamo-resource, not table-resource.
            return self._paginate_all_items_generator(
                method='batch_get_item',
                params=copy_params,
                use_table=False
            )

        # Go though all the keys and grab them 100 at a time from Dynamo.
        # Dynamo only supports a max of 100 keys at a time when doing a 'batch_get_item'.
        items_requested = []
        have_range = bool(range_name)

        uniquified_keys = keys
        if not isinstance(keys, set):
            uniquified_keys = set(uniquified_keys)
        uniquified_keys = list(uniquified_keys)

        for i in range(0, len(uniquified_keys), 100):
            key_subset = [key.key_as_dict() for key in set(uniquified_keys[i:i + 100])]
            for x in batch_pagination_generator(list(key_subset)):
                yield x

    def _parse_keys_from_query(self, query: Query) -> Optional[List[DynKey]]:
        query = _ProcessedQuery.process_query(query, api=self.api)

    def _parse_id(
            self, _id: Union[str, Iterable[str]]
    ) -> List[DynKey]:
        keys = []
        if not _id:
            return keys

        api = self.api
        for current_id in xloop(_id, iterate_dicts=True):
            keys.append(DynKey(api=api, id=current_id))

        if not keys:
            return keys

        return keys

    def query(self, query: Query = None, **dynamo_params: DynParams) -> Iterable[M]:
        """
        Forces `DynClient` to use a query. If you want a way for client to automatically
        figure out the best way to execute your query, use one of these instead:

        - `DynApi.get`
        - `DynClient.get`

        For more info see:

        - [Advanced Queries](#advanced-queries)
            - [Examples](#examples_1)

        For a quick summary on how to provide query,
        see 'query' argument doc (just a few lines down).
        But I would highly recommend looking at [Advanced Queries](#advanced-queries) for
        more details with examples!

        Args:
            api (DynApi): BaseApi object to use, this is how we know the table name, model class,
                etc.
            query (Query): You can give a simple dict here, modeled after how the standard rest-api
                query dict's work. Dict keys are the attribute/key names and values are what
                to filter on for that name.  Operators after double `__` work just like you
                would expect for our xyngular API's... here is an example:

                ```python
                { "some_attr_name__gte": 2 }
                ```

                In this case, we look for `some_attr_name` greater than or equal to integer `2`.

                You MUST provide at least one value for the hash key of the table when using
                `query` or boto3/dynamo will raise an exception.

                This is an easy way to fill out `KeyConditionExpression` and/or `FilterExpression`.
                This method can figure out which attribute goes with which one and construct
                both expressions as needed.

                For more information with examples see [Advanced Queries](#advanced-queries).

            **params (DynParams): You can provide other standard boto3 query parameters here as you
                need. If you provide both dynamo_params and query, the ones in query will overwrite
                ones in dynamo_params if there is a conflict;

                The `query` param could use either a `KeyConditionExpression` or `FilterExpression`
                depending on what attributes are in the query dict.

        Yields:
            M: (DynModel subclass instances) - The next object we got from dynamo.
                This method returns a generator that will eventually go through all the results
                from dynamo in a memory-efficient manner.
        """

        # todo: support in/lists as values....
        # todo: support `id`.

        structure = self.api.structure
        hash_key = structure.dyn_hash_key_name
        query = _ProcessedQuery.process_query(query, api=self.api)

        # if 'id' in query:
        #
        #
        # hash_key in query:

        # 1. If we have 'id', iterate though that and get DynKey's out of them
        # 2. Look at hash/range keys and try to match them up if they are lists into DynKey's
        # 3. Considering auto-finding out if we have a list of keys and can just do batch-get

        # Query for each dyn-key we find.
        keys = query.dyn_keys()
        if not keys:
            raise XRemoteError(
                "query got called with a query that had no valid DynKey's in it. "
                "This means we could not find any part(s) of the primary key we could use to do "
                "a query on (a Dynamo query requires at least a hash-key). "
                ""
                "If you have no conditions and just want to "
                "simply retrieve every item in the table use `DynClient.get` with no parameters. "
                ""
                "If you do have conditions/filters in query you need to do a `DynClient.dyn_scan` "
                "and have Dynamo scan the entire table. This will allow dynamo to evaluate your "
                "conditions on every item in the table."
            )

        for dyn_key in query.dyn_keys():
            params = {**dynamo_params}
            self._add_conditions_from_query(
                query=query,
                params=params,
                dyn_key=dyn_key
            )

            for value in self._paginate_all_items_generator(method='query', params=params):
                yield value

    def scan(self, query: Query = None, **dynamo_params: DynParams) -> Iterable[M]:
        params = {**dynamo_params}
        self._add_conditions_from_query(
            query=query,
            params=params
        )
        return self._paginate_all_items_generator(method='scan', params=params)

    def _add_conditions_from_query(
            self,
            query: Query,
            params: DynParams,
            dyn_key: DynKey = None,
            filter_key: str = 'FilterExpression'
    ):
        if not query and not dyn_key:
            return

        query = _ProcessedQuery.process_query(query, api=self.api)
        key_names: Set[str] = set()
        api = self.api
        structure = self.api.structure

        if dyn_key:
            key_names.add('id')
            key_names.add(structure.dyn_hash_key_name)
            range_name = structure.dyn_range_key_name
            if range_name:
                key_names.add(range_name)

        def add_criterion(cond_list, condition_base, name, operator, value):
            # It just so happens the basic Django filter operators are generally named the same
            # as the ones in the boto3 dynamo library. So we grab the condition/operator
            # via the same names. _ProcessedQuery will normalize the names for us.

            # exists/not_exists don't require a 'value' parameter,
            # so we need to interpret/parse query value ourselves and do the right thing.
            operator_needs_param = True
            if operator == 'exists':
                operator_needs_param = False
                if not bool_value(value):
                    # False value, so swap to the inverse/opposite operator.
                    operator = 'not_exists'
            elif operator == 'not_exists':
                operator_needs_param = False
                if not bool_value(value):
                    # False value, so swap to the inverse/opposite operator.
                    operator = 'exists'

            # Construct condition by allocating base, grabbing operator and assigning value.
            operator = getattr(condition_base(name), operator, None)
            condition = None
            if operator:
                field = structure.get_field(name)
                if operator_needs_param and field and field.converter:
                    value = field.converter(
                        api,
                        Converter.Direction.to_json,
                        field,
                        value
                    )

                operator_params = [value] if operator_needs_param else []
                condition = operator(*operator_params)

            # If we found a condition operator, use it.
            # Otherwise, we construct and raise a helpful error message.
            if operator is not None:
                cond_list.append(condition)
                return

            # Get all available conditions/operators from boto3 class so we can list them
            # in the exception message.
            available = [
                f for f in dir(condition_base)
                if callable(getattr(condition_base, f)) and not f.startswith("__")
            ]
            supplemental_msg = ""
            if condition_base is conditions.Key:
                supplemental_msg = (
                    f"Attr ({name}) is part of primary key, there are reduced "
                    f"operators available for keys when using a query. "
                    f"We could route this to a 'scan' operation, that would work.... "
                    f"Right now we don't automatically route this to a 'scan' operation "
                    f"because that's much slower and you probably really are wanting "
                    f"to do a query. You can use `dyn_scan` your self directly if that's "
                    f"what you really want to do. Or we could implement an Option/Flag to "
                    f"auto-route to a scan operation when needed."
                )

            raise XRemoteError(
                f"Using unknown boto3/dynamo operator ({operator}), "
                f"for query on attr ({name}); "
                f"the available ones are ({available}). "
                f"{supplemental_msg}"
            )

        filters = []
        keys = []

        for (name, criterion) in query.items():
            if name in key_names:
                # This is handled later...
                continue
            for (operator, value) in criterion.items():
                add_criterion(
                    cond_list=filters,
                    condition_base=conditions.Attr,
                    name=name,
                    operator=operator,
                    value=value
                )

        # Add the dyn-key conditions if needed...
        if dyn_key:
            add_criterion(
                cond_list=keys,
                condition_base=conditions.Key,
                name=structure.dyn_hash_key_name,
                operator='eq',
                value=dyn_key.hash_key
            )

            range_key = dyn_key.range_key
            if range_key:
                operator = dyn_key.range_operator or 'eq'
                # If we have an 'in' operator, we translate that to 'eq' for this purpose.
                # We should be called with separate values if there is more than one dyn_key,
                # and so are 'simulating' the `is_in` operator aspect.
                if operator == 'is_in':
                    operator = 'eq'
                add_criterion(
                    cond_list=keys,
                    condition_base=conditions.Key,
                    name=structure.dyn_range_key_name,
                    operator=operator,
                    value=range_key
                )

        params_to_mod = (('KeyConditionExpression', keys), (filter_key, filters))
        for (param_key, exp_list) in params_to_mod:
            for key in exp_list:
                exp = params.get(param_key)
                exp = exp & key if exp is not None else key
                params[param_key] = exp

    def _table_or_batch_writer(self) -> Union[BatchWriter, TableResource]:
        """
        Gets either a table or a batch-writer. So you should only call methods that are
        supported by a BatchWriter on this [since it could be one].
        All BatchWriter methods are also supported by a Dynamo TableResource so you'll be safe
        as long as you limit calls to what BatchWriter supports.
        """
        batch_writer = _DynBatchResource.grab().current_writer()
        if batch_writer:
            return batch_writer.batch_writer(api=self.api)

        return self.api.table

    def _get_item_by_id(self, _id: str) -> Optional[M]:
        api = self.api
        structure = api.structure

        if structure.dyn_range_key_name:
            split_id = _id.split(structure.dyn_id_delimiter)
            if len(split_id) != 2:
                raise XRemoteError(
                    f"Have id ({_id}) but delimiter ({structure.dyn_id_delimiter}) is in it "
                    f"more/less than once and {api} needs both a hash and range key."
                )
            hash_key, range_key = split_id
        else:
            hash_key = _id
            range_key = None

        # Construct an expression with hash and if needed, range keys.
        key_exp = conditions.Key(structure.dyn_hash_key_name).eq(hash_key)
        if range_key:
            key_exp = key_exp & conditions.Key(structure.dyn_range_key_name).eq(range_key)

        response = api.table.query(
            ConsistentRead=True,
            KeyConditionExpression=key_exp
        )

        data: Optional[JsonDict] = None
        db_datas = response['Items']
        if db_datas:
            data = db_datas[0]

        if not data:
            return None

        return api.model_type(data)

    # todo: BaseModel objects are capable of letting us know if something actually changed or not.
    #       At some point take advantage of that.
    #       This would allow us to prevent putting an unchanged item into dynamo [saves cost].
    def _put_item(self, item: 'DynModel', condition: Query = None):
        """
        Put item into dynamo-table. WON'T use any current batch-writer if a condition is supplied.
        If a condition is supplied, we always use the table and execute put immediately.

        Object will only be sent if there are any changes in object
        (compared to what was originally retrieved from table).

        If there is a change, entire object with all current attributes will be sent
        (a put fully replaces the item in table, it's not a PATCH).

        Condition will be sent with put if provided, it will be checked against anything
        that is currently in table before it's replaced in a transaction-safe way.

        Args:
            item: Item to put into dynamo table; if condition not supplied we check for a current
                batch-writer resource and use that if there is one.
            condition: Conditional query; query is sent to dynamodb, and it will check condition
                against any existing item in a transaction-safe way. If condition checks out
                then the put is done.  Otherwise, it won't be.
                If it is not, we add a field-error to object to indicate conditional check failed:

                `item.api.response_state.add_field_error('_conditional_check', 'failed')`
                or you can use `xyn_model_dynamo.const.CONDITIONAL_CHECK_FAILED_KEY` for
                the field key.
        """
        # Check to see if there is anything I actually need to send.
        if not item.api.json(only_include_changes=True, log_output=True, include_removals=True):
            log.info(f"Dynamo - {item} did not have any changes to send, skipping.")
            return

        structure = self.api.structure
        hash_name = structure.dyn_hash_key_name
        if not getattr(item, hash_name, None):
            raise XRemoteError(f"Item {item} needs a value for hash key ({hash_name}).")

        range_name = structure.dyn_range_key_name
        if range_name and not getattr(item, range_name, None):
            raise XRemoteError(f"Item {item} needs a value for range key ({range_name}).")

        # To keep things simple, I am using 'put' which replaces entire item,
        # so get all properties of item regardless if they changed or not.
        # todo: Check for primary key and raise a nicer, higher-level exception in that case.
        item.api.response_state.reset()

        params = {
            "Item": item.api.json()
        }

        resource = self._table_or_batch_writer()
        if condition:

            # Can't batch-put things with conditions, so we do them one at a time instead.
            resource = self.api.table

            # Add the conditional query to the dynamodb params dict...
            self._add_conditions_from_query(
                query=condition, params=params, filter_key='ConditionExpression'
            )

        # Finally, tell the boto resource to put the item:
        try:
            resource.put_item(**params)
        except dynamodb.meta.client.exceptions.ConditionalCheckFailedException as e:
            # By default, we ignore conditional check failed, because this means we did not want
            # to delete the item on purpose; ie: it's not an error.
            # conditions are mostly used to prevent race-conditions, hence ignored by default.
            # If someone in the future ends up wanting the exception, we can have an argument
            # or some other way to indicate that we should re-raise the condition check exception.
            # See `self.delete_obj` doc comment (above) for more details.
            log.info(
                f"Didn't send/put dynamo obj ({item}) due to condition not met ({condition}), "
                f"exception from dynamo ({e})."
            )
            state = item.api.response_state
            state.had_error = True
            state.errors = [
                'The conditional check failed, item was not deleted.',
                {'conditional-check-failed-exception': e}
            ]
            state.add_field_error(field=CONDITIONAL_CHECK_FAILED_KEY, code='failed')

    def _get_all_items(self):
        return self._paginate_all_items_generator(method='scan', params={})

    def _paginate_all_items_generator(
            self, method: str,
            params: Dict[str, Any],
            use_table=True,
    ) -> Iterable[M]:
        api = self.api
        model_type = api.model_type
        # Get table name, and also ensures table exists.
        table = api.table
        table_name = table.name
        resource = table if use_table else DynamoDB.grab().db

        while True:
            table_method = getattr(resource, method)
            response = table_method(**params)
            last_key = response.get('LastEvaluatedKey', None)

            db_datas = response.get('Items')

            if not db_datas:
                responses: Dict[str, List] = response.get("Responses")
                if responses:
                    db_datas = responses[table_name]
                else:
                    db_datas = tuple()

            for data in db_datas:
                yield model_type(data)

            if last_key:
                params['ExclusiveStartKey'] = last_key
                continue

            unprocessed = response.get('UnprocessedKeys')
            if not unprocessed:
                return

            # We need to try the fetch again for the remaining items...
            # todo: Boto/AWS recommend an exponential backoff when we retry in this case...
            #   see BatchGetItem / batch_get_item
            params['RequestItems'] = unprocessed
            continue

    def create_table(self):
        """
        This is mainly here to create table when mocking aws for unit tests. If the table
        really does not exist in reality, this also can create it.

        Time To Live Notes:

        We can't enable TimeToLive during table creation, we have to wait until after it's
        created. This is a minor issue, since the table will still function correctly,
        the queries will still filter out expired items like normal.  The only difference
        is we could get charged extra for storage we are not using. We need to still filter
        items out of our queries because deletion does not happen immediately [could take up
        to 48 hours].

        At this point, it's expected that you'll have to go into the AWS dynamo console
        to setup automatic TimeToLive item deletion for a table if you want this to create it for
        you.

        In reality, serverless framework is expected to setup the real tables for services that
        are running directly in aws; and that's where you should setup the TTL stuff for real
        tables.
        """
        structure = self.api.structure
        hash_key = structure.dyn_hash_key_name
        hash_type = structure.get_field(hash_key).type_hint
        key_schemas = [
            # Partition Key
            {'AttributeName': hash_key, 'KeyType': 'HASH'}
        ]
        attribute_definitions = [
            {
                'AttributeName': hash_key,
                'AttributeType': get_dynamo_type_from_python_type(hash_type)
            }
        ]

        # If we have a range-key, add that in.
        range_key = structure.dyn_range_key_name
        if range_key:
            range_type = structure.get_field(range_key).type_hint
            key_schemas.append({
                'AttributeName': range_key,
                'KeyType': 'RANGE'
            })
            attribute_definitions.append({
                'AttributeName': range_key,
                'AttributeType': get_dynamo_type_from_python_type(range_type)
            })

        return DynamoDB.grab().db.create_table(
            TableName=structure.fully_qualified_table_name(),
            KeySchema=key_schemas,
            AttributeDefinitions=attribute_definitions,
            BillingMode='PAY_PER_REQUEST',
            Tags=[{'Key': 'DDBTableGroupKey', 'Value': xcon_settings.service}]
        )



"""
## ORM Dynamo Overview
[orm-dynamo-overview]: #orm-library-overview

Intended use of this library is for an quick way to get/retrieve objects from Dynamo tables.

Things that it can help with:

- If table does not exist, it will create it automatically.
    - Helps when unit-testing with `@moto.mock_dynamodb` decorator, since mock expects you to
        create the table before using it. This means tables are automatically/lazily created as
        needed (with no special effort on the part of the unit test).
        .. tip:: To use
    - When running any code locally, it can automatically create the table for you without
        any special effort.
    - Normally, lambdas are not given permission to create table as we want
        cloud-formation/serverless to manage the table.
        During deployment into aws serverless/cloud-formation should be the one creating table
        if needed
- Works like with the other standard `xmodel.remote.model.RemoteModel`'s:
    - `DynModel` and other related classed work very similar.
        Supports the same basic/common methods.
    - Easy to use single 'string' to identify any dynamo object via `DynModel.id`
        - will synthesize a string representing the full primary key, for use
            with other BaseModel's when using child objects (ie: they know how to look each other up).
    - Easy/Standard way to Paginate
        - When you use the standard `DynApi.get` method, it will return a generator giving you
            one object at a time while bulk-getting up to 100 per-request to dynamo, depending on
            how it had to query the data for you
            (the goal is for it to figure out the most efficient way to query automatically).
        - You don't have to worry about what page of results you are on or how it works,
            you just loop/run the generator and it will eventually give you back all of
            the objects. [Auto Prefetch Children](./#auto-prefetch-children)
        - **TODO** (rest-client `xmodel_rest.RestClient` has this, need to do it still for
            `DynClient`).  You can still use methods in `xmodel.children`
            to bulk-grab them.
            - If your curious, see [Auto Prefetch Children](./#auto-prefetch-children)
                for more info about how to use the auto-pre-fetch feature with the standard
                `xmodel.base.model.BaseModel`.
    - Simple way to insert/update/delete objects individually or in bulk.
    - Full power of the JSON to/from BaseModel's infrastructure.
        - Including automated conversion of types to/from, such as dates.
        - Default values
        - Read only fields, etc.
    - Central spot to put future high-level dynamo code to share among our projects.

### Quick Start

If you don't know much about the ORM, have a look at
[ORM Library Overview](./#orm-library-overview) first.  It's an overview of the basic concepts.

Index/Summary of the main classes you'll be interacting with:

- `DynModel`: Represents an object in a table.
- `DynApi`: Basically represents the table, it's the central 'hub' class that lets you get to the
    - `DynStructure`: List of fields and other class-level info about the `DynModel`.
    - `DynClient`: Wraps boto, figures out the request to use with boto and executes it.
- `DynKey`: Contains a hash + range keys, along with ways to put them together into an
    `DynModel.id` string and splitting them apart again.
- `DynField`: Represents a field on a `DynModel`.  Automatically created in `DynStructure` if
    it's not user allocated on a attribute/field on `DynModel`.
    - `HashField`: Special `DynField` object, indicated field is the Hash of the table.
    - `RangeField`: Indicates the Range field of the table (if there is one).
- `DynBatch`: Context manager (ie: `with` object).
    Allows you to batch non-transaction put's (so system will just use strait put') and deletes.

### Example Data Models

Examples are probably the best way to get a 'quick start', here are some below.

First, I'll get a few BaseModel's defined. After I'll show examples of using them.

This first one show's a table with a Hash + Range key, along with a list of dicts along
with some basic data fields (str/bool).

.. note:: You can see models vary similar to these in action in sine unit-tests
    Look at
    [tests/test_dynamo.py](https://github.com/xyngular/py-xyn-model-dynamo/blob/master/tests/test_dynamo.py)
    in xdynamo source if your interested.

>>> class ModelWithRangeKey(
...     DynModel,
...     # ---> used for end of table name:
...     dyn_name="modelWithRangeKey"
... ):
...     my_hash: str = HashField()
...     my_range: str = RangeField()
...     name: str
...     a_number: int
...     hello: bool
...     items: List[Dict[str, str]]

Here is a normal non-dynamo model. We will be using this as a way to parse a sub-dict
automatically into a regular model object.
We enabled `Field.include_in_repr` in the below example, it will make `sub_name` print out
in string when object is converted to a string (such as when logging object out).

>>> from xmodel.fields import Field
>>> class ModelAsSubJsonDict(BaseModel, has_id_field=False):
...
...     # It puts 'sub-name' into the object description when converting object
...     # to a string
...     # (ie: such as when you log out a object of type 'ModelAsSubJsonDict')
...     sub_name: str = Field(include_in_repr=True)
...     queue: bool

Here is a second `DynModel` for a separate table.  It has a relationship to a
`ModelWithRangeKey`.

>>> class ModelOnlyHash(
...     DynModel,
...     dyn_name="visibleShipConfirm",
...     dyn_service="experimental"
... ):
...     hash_only: str = HashField()
...     name: str
...     items: List[Dict[str, str]]
...     a_number: int
...     test_item_id: str
...     test_item: ModelWithRangeKey
...     sub_item: ModelAsSubJsonDict

In the real dynamo table, it would store `VisiblePackage.sub_item` as a
`test_item_id` attribute by grabbing the `DynModel.id` from the `ModelWithRangeKey` object.

It would lazily lookup object if you try to access `VisiblePackage.sub_item`, just like you
would expect.

.. todo:: rest-client `xmodel_rest.RestClient` has an ability to auto pre-fetch children.
    Still need to do it for `DynClient`. You can still use methods in `xmodel.children`
    to bulk-grab them.

    If your curious, see [Auto Prefetch Children](./#auto-prefetch-children)
    for more info about how to use the auto-pre-fetch feature with the standard
    `xmodel.base.model.BaseModel`.

Create a few items in a table (for illustrative purposes, for the following examples):

>>> ModelWithRangeKey(my_hash="my-h1", my_range="my-r1", name="A").api.send()
>>> ModelWithRangeKey(my_hash="my-h2", my_range="my-r2", name="B").api.send()

### Basics of Getting Items

Quick Example of getting an item, this gets the item by a `"my-h1"` hash-key and
`"my-r1"` range-key:

>>> ModelWithRangeKey.api.get_via_id({'my_hash': "my-h1", 'my_range': "my-r1"})
ModelWithRangeKey(my_hash: "my-h1", my_range: "my-r1")

Various ways Get Item are below, in general you need a hash-key to do a dynamo-query and to
generally get items. If you don't have a hash-key value, then you must scan the table.

Right now, you can grab all items in a table via `DynApi.get`.
This will pass a blank query to `DynClient.get`, which will make it do a
full-table scan and return all items.

Right now, `DynClient` won't scan and will instead raise an
`xmodel.remote.errors.XRemoteError`
if you pass in a non-blank query without including a hash-key value.

.. todo:: Support scanning entire table with a non-blank query to filter it with.
    This will be supported in the future, right now it's unimplemented.
    Most of the time, you'll really, really want to query the table in any case.
    Querying the table (with a hash-key) is **MUCH** faster then scanning it.

I have various examples below, but in general we support querying in these ways,
you can use `DynApi.get` or `DynClient.get`.
These methods will figure out the best way to execute get request/query, generally detailed
below:

- If provided query is blank, scans and returns all items/objects in table.
- Generally if you have both (and only) hash and range keys, it will do a batch-get automatically
    which lets us query for 100 objects at a time. Otherwise it will fall-back to a query.
    When doing a query, dynamo only supports one query-request per-hash.
    - If the table structure has only a hash-key (and no range key),
        then it's a list of only hashes.
- If you only have a:
    - hash-key:  Will return all objects with that hash, could be multiple objects if the table
        structure has a range-key.
        - If table only supports hash-key, it will be either a single object or empty list.
    - single hash + range keys: Return's a single object if it exists, or empty-list.
        - Consider using `XynApi.get_via_id`, it won't return a `list` if you provide
            the `id` value as a single string/dict as first argument.
    - multiple Hash + range keys: Will query every combination of hash + range keys automatically.
        It can use a batch-get for this and attempt to lookup up to 100 objects per-request.
        If you have more than 100, we will split up the requests for you automatically
        (you'll just see a single stream of objects come back).
    - `id`: If you have the `DynModel.id` for an object, it contains one or both keys and can
        be used to query object via `DynApi.get_via_id`.
    - `DynKey`: You can use this via `DynApi.get_via_id` to query for the object.
        - `DynKey`'s represet one or both components of a DynamoDB primary key.
- If you have other attributes besides just the range/hash key:
    - We need to fallback to a query in this case, to support filtering by non-key attributes.
    - One query per-hash/range key in the query.
        - If the range-key is using `between` operator, we still only do one query since dynamo supports
            this operator in a query.
            - We'll talk more about operators later in the [Advanced Queries](#advanced-queries)
                section further on.
        - If you use a list of values with range-key, it will have to use
            multiple queries, one per-hash/range key combination(s) provided.


First there is `DynApi.get_via_id` which can take a list of id strings,
or a list of dicts with a hash-key and (optionally) a range-key:

>>> ModelWithRangeKey.api.get_via_id({'my_hash': "my-h1", 'my_range': "my-r1"})
ModelWithRangeKey(my_hash: "my-h1", my_range: "my-r1")

This will produce an error, you must provide all parts of the key to use `get_via_id`,
so it needs the range-key part:

>>> ModelWithRangeKey.api.get_via_id({'my_hash': "my-h1"})
Raises XRemoteError

If you want all the objects for a particular hash regadless of the range-key, you can
uss the `DynApi.get` method instead. We wrap it in a list, because a generator is
normally returned (to showcase all output). You can put the returned generator in
a `for` loop instead if you want. The generator will correctly paginate all results
for you automatically. Here is the example:

>>> list(ModelWithRangeKey.api.get({'my_hash': "my-h1"}))
[ModelWithRangeKey(..., my_range="my-r1"), ModelWithRangeKey(..., my_range="my-r2")]

You can also query on other non-key attributes,

>>> list(ModelWithRangeKey.api.get({'my_hash': "my-h1"}))

You can also use a generic object that represents a dynamo key with `DynKey`.
You can pass `DynKey` objects directly into `DynApi.get_via_id`.
Other methods let you also directly pass `DynKey`'s, such as `DynClient.delete_objs`.

>>> key = DynKey(hash_key='my-h1', range_key='my-r1')
>>> ModelWithRangeKey.api.get_via_id(key)

To you can grab a list of them, a list of dicts or a list of keys,
when you provide a list you get a generator back that will paginate though
the results correctly.

>>> list(ModelWithRangeKey.api.get_via_id([key]))
[ModelWithRangeKey(....)]

Also, if your table only has a hash-key, you can just directly provide it's string/int value;
or a list of them:

>>> ModelOnlyHash.api.get_via_id('a-hash-key')

This is the same value as you get from the model's `DynModel.id` attribute.

>>> obj = ModelOnlyHash(hash_only = 'a-hash-key')
>>> assert obj.id == 'a-hash-key'

Objects with a range-key by default have them joined together with hash-key via a pipe `|`
delimiter, like so:

>>> obj = ModelWithRangeKey(my_hash="my-h1", my_range="my-r1")
>>> assert obj.id == 'my-h1|my-r1'

You can change the joining/delimiter string via `DynStructure.dyn_id_delimiter`.
Just like with the normal `xmodel.base.model.BaseModel`, you can set these attributes via the
class arguments on `DynModel` subclasses.
(for more details on class arguments, see `xmodel.base.model.BaseModel.__init_subclass__`
and `DynStructure.configure_for_model_type` for the DynModel specific ones available).

Every `DynModel` has this 'virtual' `DynModel.id` value that is the primary key of
the object. THis is the hash-key, plus range-key value (if object has one) of an object.
This single string uniquely identified the object.

The rest of the `xmodel.remote` can use this just like any `id` from
other `xmodel.base.model.BaseModel` objects. Meaning, this virtual `id` value can be
stored in other places to form relationships (as you can see with the above
example on `ModelOnlyHash.sub_item`.



### Updating Items

Just like other `xmodel.base.model.BaseModel` objects, `DynModel` can be changes and then
the changes sent to Dynamo like so:

>>> obj = ModelOnlyHash(hash_only="a-hash-key")
>>> obj.carrier = "new-carrier"
>>> obj.api.send()

You can also mass-update/create objects via `DynClient.send_objs`:

>>> list_of_objs: List[ModelOnlyHash]
>>> ModelOnlyHash.api.client.send_objs(list_of_objs)

Right now we only support "Putting" objects into dynamo (ie: not patching them).
It will replace the entire item in dynamo, no mater what you change or update on the object.

.. todo:: In the future, there will be an option to patch object(s) via `DynTransaction`.
    The feature has not been finished yet.
    It will use a Dynamo Transaction to do it.
    See [Todo/Future: Batch via Transaction](#todofuture-batch-via-transaction)

### Deleting Items

>>> obj.api.delete()

You can mass delete via:

>>> key = DynKey(hash_key='my-h1', range_key='my-r1')
>>> objs = list(ModelWithRangeKey.api.get_via_id(key))
>>> ModelWithRangeKey.api.client.delete_objs(objs)

You can also give the `DynClient.delete_objs` a list of `DynKey`'s.

>>> ModelWithRangeKey.api.client.delete_objs([key])

This allows you to delete objects without having to create full-models and looking to see
which field is the hash/range key.  The `DynKey` always accepts the values the same way.


### Batch Updating / Deleting

You can send a one-off list of multiple objects to update/delete all at once (see examples above).

If you want a section of code that gets executed to batch-delete/update you can directly use
the `DynBatch` class. It's a context-manager and can apply batch Put's and Delete's via a
`with` statment that will continue to apply it no mater how deep the call stack is inside
the `with` statement. See `DynBatch` for more details

Quick Example:

>>> # DynModel objects of some sort....
>>> obj1: ModelOnlyHash
>>> obj2: ModelOnlyHash
>>> obj3: ModelOnlyHash
>>> with DynBatch():
...     obj1.carrier = "changed"
...     obj1.api.send()
...     obj2.carrier = "changed"
...     obj2.api.send()
...     obj3.api.delete()

This would end up sending both updates and the delete in the same request.
It works by 'batching' a number of objects at a time and sending them.
If there are still objects to send by the time the `with` statement exits,
the renaming unsent objects are sent to dynamo.

### Todo/Future: Batch via Transaction

.. todo:: In the future, there will be a class called `DynTransaction` that you can use to
    batch transactions together. It would also allow bulk-partial updating/patching objects
    instead of using 'puts' to replace entire object. And to tie a set of objects together
    that must all be written or rolled back in one go.

## Mocking Dynamo with moto

You can use the `moto` 3rd part dependency, and their `@moto.mock_dynamodb` method decorator
to 'mock' dynamo. There is something to be aware though, but they are easy to get correct:

- The mocking library expects you to first your dynamo tables via boto3 calls
    before they are used.  `xdynamo.DynClient` will check and ensure the tables
    are created automatically in a lazy fashion
    (ie: first time an attempt to send/get a `DynMode` from Dynamo).
- In an effort to reuse already opened connections to dynamo, we use a shared resource
    `xdynamo.dyn_connections.DynamoDB`.  It's important that the connection is created
    while `moto` is active for `moto` to mock the connection/session.

    .. note:: `xinject.fixtures.context` fixture is automatically used when `xinject` is
        installed as a dependency.
        This ensures when a `xdynamo.dyn_connections.DynamoDB` is asked for, a brand new
        one will be created for the unit test and therefore create a new connection/session.

    - It's relatively expensive to open a new connection vs using an already existing
        connection to Dynamo since Dynamo has to setup a encryption key context in it's self.

## Advanced Queries
[advanced-queries]: #advanced-queries

Ensure you've read the earilier segment about [Basics of Getting Items](#basics-of-getting-items)
first before reading this more advanced section.

`DynApi.get` and `DynClient.get`, for the `query` param can accept more then
just the hash-key and range-key; other attributes are allowed to filter the results even more!

Dynamo has a concept of a 'query' and a 'scan' to get data. A query needs to at least each
by the hash-key (range-key is optional).  Scan will let you search with any attribute but has
to look at all records, and so is slow.

If you at least provide the hash-key, you can filter by other attributes and still get
decent performance (as it only has to scan records with that hash).  If you also provide the
range-key, that's even better as that will narrow it down to one-record, and is fast since
Dynamo can use an index.

You can also provide a list of hash-keys, with other attributes to fitler by.
`DynClient` will automatically break it up into multiple queries, one per-hash key
automatically and paginate all the results together for you.

### Operators

It also supports operators, like `some_field__gte: 3` would look for things that are
greater-than-or-equal to `3`. A hash-key must be exact, or a list (ie: `__in`).

When wanting to query, the hash-field has to be an exact-equals value (no other special operators)
or a `list` value and/or `__in` operator.  When it's a list (which maps it's self automatically
to the `__in` operator by default) we simulate it by doing multiple queries, one per-item
in the list for the hash and then concatenating the results together via the generator
that gets returned (ie: lazily execute each query as the generator runs).

To get full list of operators, see operator method names on `bot3.dynamodb.conditions.Attr`
(and it's AttributeBase superclass). Also have a mapping at `operator_alias_map` that maps
some of our standard xyngular-api operators to how it's named in Dynamo.

I also have a list right below:

non-hash query operators you can use:

#### Operator List

- eq
- lt
- lte
- gt
- gte
- begins_with
- between
- is_int
_ exists
_ not_exists
- contains
- size
- attribute_type

See examples below to see how operators can be used.

For more info, take a look at some of the following examples and also look at
`DynClient.get`.

### Examples
[examples_1]: #examples_1

```python
ModelOnlyHash.api.get(
    query={
        'hash_only':['vis-track-id-2', 'vis-track-id-1'],
        'carrier': 'c2'
    }
)
```

This will do two queries in Dynamo, one for each:

1. 'hash_only: vis-track-id-2' + 'carrier: c2'
2. 'hash_only: vis-track-id-1' + 'carrier: c2'

It will combine the results of both together into a single generator and return the generator.
You can iterate on the generator to get all the results, all paginated for you.

You can use operators after the keyword, just like you can do for the Xyngular APIs:

```python
ModelOnlyHash.api.get(
    query={
        'hash_only':['vis-track-id-2', 'vis-track-id-1'],
        'a_number__gte': 2
    }
)
```

Looks for the two different `hash_only`'s that have a `a_number` attribute that
is greater then or equal to 2.


>>> my_objs = ModelWithRangeKey.api.get(
...     query={
...         'my_hash':['hash-1', 'hash-2'],
...         'my_range':['range-1', 'range-2'],
...         'a_number__gte': 2
...     }
... )


It looks for all of the following combinations of range/hash with four seperate queries
(internally):

- hash-1, range-1
- hash-1, range-2
- hash-2, range-1
- hash-2, range-2

Each of these queries will also have a filter where `a_number` attribute will need
to be greater then or equal to 2.

`DynApi.get` will return a generator, that will return all objects returns from
these four internal queries it will do for you.

All you have to do is run the generator to see all the objects, like so:

>>> for o in my_objs:
...     print(o.a_number)
--- outputs one number per-object in table that match query ---

If the above example query only had `my_hash` and `my_range` and **NOT** `a_number__gte`,
it would have executed a very-fast batch-get on dynamo with the hash/range key combinations.

Doing four query/requests are still pretty fast,
but doing a batch-get can be executed in a single request (100 key combinations per-request)!

Like I said earlier, the nice thing about `DynApi.get` / `DynClient.get`
is that it can figure out the best way to query the objects for you based on what you give
so you don't have figure it out your self.

"""  # noqa -- Some lines can't be cut shorter (Url's).

from xdynamo.model import DynModel
from xdynamo.fields import DynField, HashField, RangeField
from xdynamo.common_types import DynKey
from xdynamo.resources import DynBatch

__all__ = [
    "DynKey", "DynModel", "HashField", "RangeField", "DynBatch", "DynField"
]



from typing import TypeVar, Optional, Any, Union, List, Dict, Iterable, TYPE_CHECKING
from uuid import UUID

from boto3.dynamodb.table import TableResource

from .db import DynamoDB

from xmodel.common.types import FieldNames
from xmodel.remote import XRemoteError
from xmodel.remote.api import RemoteApi
from xdynamo.client import DynClient
from xdynamo.common_types import DynKey
from xdynamo.fields import DynField
from xdynamo.structure import DynStructure
from xsentinels.default import Default
from xurls.url import Query
from xloop import xloop


if TYPE_CHECKING:
    from .model import DynModel
    M = TypeVar('M', bound=DynModel)
else:
    M = TypeVar('M')


def lazy_load_types_for_dyn_api(cls):
    """
    Lazy load our circular reference right before it's needed.
    This is put in as DynModel's lazy_loader, xyn_model will call us here when needed.

    See `xyn_model.base.model.BaseModel.__init_subclass__` and it's lazy_loader argument
    for more details.
    """
    if 'DynModel' not in globals():
        from xdynamo.model import DynModel
        globals()['DynModel'] = DynModel


class DynApi(RemoteApi[M]):
    """
    Put things here that are only relevant for all DynModel's.

    The main change vs the base BaseApi class is filtering out of the JSON invalid blank values.

    Right now this model class is only used to transform json from/to Dynamo via
    the `json` and `update_from_json()` methods. See `xmodel.base.api.BaseApi` for more
    details.
    """
    client: DynClient[M]
    structure: DynStructure[DynField]
    # This type-hint is only for IDE, `RemoteApi` does not use it
    # (self.model_type value is passed in when RemoteApi is allocated, in __init__ method).
    model: M

    def get(
            self,
            query: Query = None,
            *,
            top: int = None,
            fields: Optional[FieldNames] = Default,
            allow_scan=False
    ) -> Iterable[M]:
        """
        Convenience method for the `self.client.get` method.

        Generally, will get/query/scan/batch-get; generally,
        the client will figure out the best way to get the items based on the provided query.
        If the query is None or blank, it will grab everything in the table.

        Args:
            query: A dict with key the field, with optional double-underscore and operation
                (such as `{'some_field__beginswith': 'search-value'}`).
                The value is what to search for.
                If you give this a list, it implies a `__in` operator, and will do an OR
                on the values in the list.
            top:
            fields:
            allow_scan: Defaults to False, which means this will raise an exception if a scan
                is required to execute your get.  Set to True to allow a scan if needed
                (it will still do a query, batch-get, etc; if it can, it only does a scan
                if there is no other choice).

                If the query is blank or None will still do a scan regardless of what you pass
                (to return all items in the table).

        Returns:

        """
        return self.client.get(query, top=top, fields=fields, allow_scan=allow_scan)

    def get_key(self, hash_key: Any, range_key: Optional[Any] = None) -> DynKey:
        """
        Easy way to generate a basic `DynKey` with hash_key, and range_key
        (if model has range key).

        If you don't provide a range-key and the model needs a range-key,
        will raise an `xmodel.remote.errors.XRemoteError`.
        """
        return DynKey(api=self, hash_key=hash_key, range_key=range_key)

    def get_via_id(
            self,
            id: Union[
                    int | str | UUID,
                    List[int | str | UUID],
                    Dict[str, str | int | UUID],
                    List[Dict[str, str | int | UUID]],
            ],
            fields: FieldNames = Default,
            id_field: str = None,
            aux_query: Query = None
    ) -> Union[Iterable[M], M, None]:
        """
        Overridden in DynApi to convert any provided `DynKey` into string-based `id` and
        passing them to super and returning the result.

        See `xmodel.remote.api.RemoteApi.get_via_id` for more details on how this method works.

        Args:
            id: In addition to `str` and `int` values, you can also used `DynKey`(s) if you wish.
            fields: See `xmodel.remote.api.RemoteApi.get_via_id`
            id_field: See `xmodel.remote.api.RemoteApi.get_via_id`
            aux_query: See `xmodel.remote.api.RemoteApi.get_via_id`

        Returns:
            See `xmodel.remote.api.RemoteApi.get_via_id`
        """
        if id is None:
            return None

        is_list = isinstance(id, list)
        if is_list:
            id: Union[DynKey, int, str]
            new_id = [v.id if type(v) is DynKey else v for v in xloop(id)]
        else:
            new_id = id.id if type(id) is DynKey else id

        return super().get_via_id(new_id, fields=fields, id_field=id_field, aux_query=aux_query)

    @property
    def table(self) -> TableResource:
        """ Returns the boto3 table resource to use for our related DynModel.
            Don't cache or hang onto this, it's already properly cached for you via the current
            Context and so will work in every situation [unit-tests, config-changes, etc]...
        """

        if not self.structure.dyn_hash_key_name:
            raise XRemoteError(
                f"While constructing {self.structure.model_cls}, found no hash-key field. "
                f"You must have at least one hash-key field."
            )

        # Look it up each time in case config/service/env/context changes enough
        # for it to be different. DynamoDB will cache the table by name and so it's
        # very fast on subsequent lookups.
        table_name = self.structure.fully_qualified_table_name()

        # noinspection PyTypeChecker
        return DynamoDB.grab().table(name=table_name, table_creator=self._create_table)

    def _create_table(self, dynamo: DynamoDB) -> TableResource:
        return self.client.create_table()

    def delete(self, *, condition: Query = None):
        """
        REQUIRES associated model object [see self.model].

        Convenience method to delete this single object in API.

        If you pass in a condition, it will be evaluated on the dynamodb-service side and the
        deletes will only happen if the condition is met.
        This can help prevent race conditions if used correctly.

        If there is a batch-writer currently in use, we will try to use that to batch the deletes.

        Keep in mind that if you pass in a `condition`, we can't use the batch write.
        We will instead send of a single-item delete request with the condition attached
        (bypassing any current batch-writer that may be in use).

        Args:
            condition: Conditions in query will be sent to dynamodb; object will only be deleted
                if the conditions match.
                See doc comments on `xyn_model_dynamo.client.DynClient.delete_obj` for more
                details.
        """

        # Normally I would call super().delete(...) and have the super have a **kwargs it can
        # simply pass along (so it can do any normal checks it does).
        # Don't want to modify another library right now, so for now copying/pasting the code here.
        model = self.model
        if model.id is None:
            raise XRemoteError(
                f"A delete was requested for an object that had no id for ({model})."
            )

        self.client.delete_obj(model, condition=condition)

    def send(self, *, condition: Query = None):
        """
        """

        # Normally I would call super().delete(...) and have the super have a **kwargs it can
        # simply pass along (so it can do any normal checks it does).
        # Don't want to modify another library right now, so for now copying/pasting the code here.
        model = self.model
        if model.id is None:
            raise XRemoteError(
                f"A send was requested for an object that had no id for ({model})."
            )

        self.client.send_objs([model], condition=condition)



from typing import Optional, Set, TypeVar, TYPE_CHECKING

from xcon import xcon_settings
from xmodel.remote import RemoteStructure
from xmodel.remote import XRemoteError
from xdynamo.fields import DynField
from xdynamo.common_types import DynKeyType
from xsentinels.default import Default

F = TypeVar('F', bound=DynField)


class DynStructure(RemoteStructure[F]):
    """
    Putting it here right now, it uses most of the super-class stuff, with a few extra fields.
    See `xmodel.base.structure.BaseStructure` for more details;

    Use these names on the DynModel's as class-arguments to configure these. Those values
    will eventually be put into this structure for use by the DynClient/DynApi classes.
    """

    # Todo: Refer to parent docs, but put more info about how virtual-id's are used for our
    #  dynamo tables.
    virtual_id: bool = True

    dyn_name: str = Default
    """ Table name, minus service/environment name.
        We generally want to use camelCase for this and no dashes/underscores.

        If left as `Default`, will use a "camelCase" version of the model class name.
    """

    dyn_service: str = Default
    """ Service to use, by `Default` we use current `xcon.conf.XconSettings.service`.
        If Blank/None we won't include it in the name.
    """

    dyn_environment: str = Default
    """ Environment to use, by `Default` we use current `xcon.conf.XconSettings.environment`.
        If Blank/None we won't include it in the name.
    """

    dyn_hash_key_name: str = None
    """ Name of hash key [for now both model and table attribute must be same name].
        If left as None, an error will be raised when we try connect to table.

        You can more easily indicate this via `HashField`, ie:

        >>> class MyModel(DynModel):
        ...     my_hash: str = HashField()

        When you do this, we will fill in `dyn_hash_key_name` for you.
    """

    dyn_range_key_name: str = None
    """ Name of range key [for now both model and table attribute must be same name];
        If left as None, we don't have one on the table.

        You can more easily indicate this via `RangeField`, ie:

        >>> class MyModel(DynModel):
        ...     my_hash: str = RangeField()

        When you do this, we will fill in `dyn_range_key_name` for you.
    """

    dyn_id_delimiter = "|"
    """
    Default way to delimit the hash/range keys when used in external systems and with code under
    the `xmodel` library. In general, it's assumed we only have one key that can uniquely
    identify model objects/items in general. For DynamoDB, if it has a range key we can merge
    both the hash-key and range-key together with this character. By default it's a pipe `|`.
    But you can set it to something else here if needed.

    This makes it so other systems can store just one value to refer to an item in a dynamo
    table, ie: "some-hash-value|some-range-value".  When other BaseModel's from other systems
    need to get a Dynamo model, they can look it up via this single 'id' value.
    """

    @property
    def dyn_hash_field(self) -> DynField:
        return self.get_field(self.dyn_hash_key_name)

    @property
    def dyn_range_field(self) -> Optional[DynField]:
        return self.get_field(self.dyn_range_key_name)

    def has_id_field(self):
        id_field = self.get_field('id')
        return (id_field.dyn_key is DynKeyType.hash) if id_field else False

    def configure_for_model_type(
            self,
            *,
            # These fields are used to name the table
            # format: `{dyn_service}-{dyn_environment}-{dyn_name}`
            dyn_name: str = Default,
            dyn_service: str = Default,
            dyn_environment: str = Default,
            dyn_id_delimiter: str = Default,

            **kwargs
    ):
        """
        See superclass method `xmodel.base.structure.BaseStructure.configure_for_model_type`
        first. It has more information about what calls this method and how to
        customize value via class arguments
        (see `xmodel.base.model.BaseModel.__init_subclass__` for more about the class argument
        angle).

        This method accepts every argument in the super-class version, with the addition
        of ones more specific to Dynamo. So se

        This method accepts named-class-parameters for `xdynamo.model.DynModel`
        subclasses.

        See super-class method `xmodel.base.structure.BaseStructure.configure_for_model_type`
        doc's for additional arguments and details.

        Args:
            dyn_name: Name of the table.  We will add dyn_service and current environment to the
                final name.  See `DynStructure.fully_qualified_table_name` for more details.

                This can also be `None` to indicate that it has no direct-table
                (ie: this is an abstract base class, with commonly shared fields, etc).

            dyn_service: Name of the service to use for that portion of the table name.
                If not provided, and `xcon` is available, will use `con.conf.XconSettings.service`.
                Right now `xcon` is a required dependency, but will make it optional in the future.

                In that future version, leaving this as `Default` would make it not format
                the service name into the table name.

            dyn_environment: Name of the environment to use for that portion of the table name.
                Right now `xcon` is a required dependency, but will make it optional in the future.

                In that future version, leaving this as `Default` would make it not format
                the environment name into the table name.

            dyn_id_delimiter: Delimiter to use for the `id` value when table has both a
                hash and range key (to delimit the values of the two).

                By default, a pipe char `|` is used.

            **kwargs: These all come from class-arguments given to the
                `xdynamo.model.DynModel` at class-definition time that need to be sent to
                my super-class via
                `xmodel.base.structure.BaseStructure.configure_for_model_type`. See that
                for more on what other arguments are supported.
        """
        super().configure_for_model_type(**kwargs)

        # Resolve default `dyn_name` if needed
        if dyn_name is Default:
            # Attempt to use model class name (with first char lower-cased) if we have one.
            model_name = self.model_cls.__name__
            dyn_name = f'{model_name[:1].lower()}{model_name[1:]}' if model_name else ''

        if dyn_name == '':
            raise XRemoteError(
                f"The `dyn_name` of dynamo model ({self.model_cls}) was blank/None; "
                f"model must be `None` or have a non-blank `dyn_name`."
            )

        self.dyn_name = dyn_name
        self.dyn_service = dyn_service
        self.dyn_environment = dyn_environment

        if dyn_id_delimiter:
            self.dyn_id_delimiter = dyn_id_delimiter

        type_to_attr_map = {
            DynKeyType.hash: 'dyn_hash_key_name',
            DynKeyType.range: 'dyn_range_key_name',
        }

        encountered_types: Set[DynKeyType] = set()
        for field in self.fields:
            # We can assume all fields are DynField subclasses, since it's the field_type
            key_type = field.dyn_key
            if not key_type:
                continue

            attr_name = type_to_attr_map[key_type]
            if key_type in encountered_types:
                last_encountered_name = getattr(self, attr_name, None)
                raise XRemoteError(
                    f"We have multiple ({DynKeyType.hash}) key-type fields: "
                    f"({last_encountered_name}) and ({field.name}) for model ({self.model_cls}). "
                    f"There can only be one ({DynKeyType.hash}) field key-type on a dynamo table."
                )

            encountered_types.add(key_type)
            setattr(self, attr_name, field.name)

        # If user manually set this to 'True', then keep it; otherwise determine it based
        # on if there was a hash-key on 'id'.
        # If the `id` field is not the hash-key, then we don't consider us having a
        # `id` field. Instead, we have a virtual-id field that is a combination of both
        # the hash-key field value with the range-key value (if we have a range-key).
        # The purpose of that is to indicate that the `id` field is not a normal field
        # on this class, but a virtual property instead (see `DynModel.id`).

        from xdynamo.model import DynModel
        if self.model_cls is DynModel:
            # Restore `DynModel.id` normal non-field property for DynModel.
            # We want it to be a virtual-id.
            # (BaseModel by default will capture field properties into `fget` and `fset`
            #  in the Field object, and then execute properties when needed after doing it's
            #  normal thing.... we want to replace that functionality with our own property,
            #  and not have it go through the normal things... but still treat 'id' like a
            #  normal property otherwise).
            id_field = self.get_field('id')
            self.model_cls.id = property(fget=id_field.fget, fset=id_field.fset)

    def fully_qualified_table_name(self) -> str:
        """
        Fully qualified name of the table in Dynamo as a str.
        Format is: '{dyn_service}-{dyn_environment}-{dyn_name}'
        """
        components = []
        service = self.dyn_service
        if service is Default:
            service = xcon_settings.service

        env = self.dyn_environment
        if env is Default:
            env = xcon_settings.environment

        if service:
            components.append(service)

        if env:
            components.append(env)

        name = self.dyn_name
        if not name:
            raise XRemoteError(
                f"Tried to get `fully_qualified_table_name` but have no table name for {self}"
            )
        components.append(name)
        return "-".join(components)

    @property
    def endpoint_description(self):
        return self.fully_qualified_table_name()



import dataclasses
from enum import Enum, auto as EnumAuto  # noqa
from typing import TYPE_CHECKING, Union, Any, Optional, Dict, Iterable, Tuple, Set, Type

from xdynamo.errors import XModelDynamoError, XModelDynamoNoHashKeyDefinedError
from xmodel.remote import XRemoteError
from xmodel.base.fields import Converter
from xurls.url import Query
from xloop import xloop

if TYPE_CHECKING:
    from xdynamo.api import DynApi
    from xdynamo.model import DynModel
    from xdynamo.fields import DynField

DynParams = Dict[str, Any]

_type_to_aws_type_map = {
    int: "N",
    float: "N",
    complex: "N",
    bool: "BOOL",
    str: "S",
    dict: "M",
    list: "L"
}

operator_alias_map = {
    "in": 'is_in',
    "exact": 'eq',
    "": 'eq',
    None: 'eq'
}
"""
Used to normalize some common operators (that we use in other systems) to the one used in Dynamo.
"""


def get_dynamo_type_from_python_type(some_type: Type) -> str:
    dyn_type = _type_to_aws_type_map.get(some_type)
    if dyn_type is not None:
        return dyn_type

    # todo: consider making the type we send to dynamo overridable
    #   default map is `get_dynamo_type_from_python_type`.
    #   generally, unless it's a basic type we default to `str`
    #   (example: datetime types use str).
    return _type_to_aws_type_map[str]


@dataclasses.dataclass(frozen=True, eq=True)
class DynKey:
    api: 'DynApi' = dataclasses.field(compare=False)
    # We only compare with `id`, this should represent our identity sufficiently.
    id: str = None
    hash_key: Union[Any] = dataclasses.field(default=None, compare=False)
    range_key: Optional[Any] = (
        dataclasses.field(default=None, compare=False)
    )
    range_operator: str = dataclasses.field(default=None, compare=False)
    require_full_key: bool = dataclasses.field(default=True, compare=False)

    @classmethod
    def via_obj(cls, obj: 'DynModel') -> 'DynKey':
        structure = obj.api.structure
        hash_name = structure.dyn_hash_key_name

        if not hash_name:
            raise XModelDynamoNoHashKeyDefinedError(
                f"While constructing {structure.model_cls}, found no hash-key field. "
                f"You must have at least one hash-key field."
            )

        hash_value = getattr(obj, hash_name)

        if hash_value is None:
            raise XModelDynamoError(
                f"Unable to get DynKey due to `None` for dynamo hash-key ({hash_value}) "
                f"on object {obj}."
            )

        range_name = structure.dyn_range_key_name
        range_value = None
        if range_name:
            range_value = getattr(obj, range_name)
            if range_value is None:
                raise XModelDynamoError(
                    f"Unable to get DynKey due to `None` for dynamo range-key ({range_name}) "
                    f"on object {obj}."
                )

        return DynKey(api=obj.api, hash_key=hash_value, range_key=range_value)

    def key_as_dict(self):
        structure = self.api.structure
        hash_field = structure.dyn_hash_field
        range_field = structure.dyn_range_field

        def run_converter(field: 'DynField', value) -> Any:
            converter = field.converter
            if not converter:
                return value

            return converter(
                self.api,
                Converter.Direction.to_json,
                field,
                value
            )

        # Append the keys for the items we want into what we will request.
        item_request = {hash_field.name: run_converter(hash_field, self.hash_key)}
        if range_field:
            item_request[range_field.name] = run_converter(range_field, self.range_key)
        return item_request

    def __post_init__(self):
        structure = self.api.structure
        delimiter = structure.dyn_id_delimiter
        range_name = structure.dyn_range_key_name
        need_range_key = bool(range_name)

        hash_key = self.hash_key
        range_key = self.range_key
        api = self.api

        _id = self.id
        if _id is not None and not isinstance(_id, str):
            # `self.id` must always be a string.
            # todo: Must check for standard converter method
            _id = str(_id)
            object.__setattr__(self, 'id', _id)

        require_full_key = self.require_full_key

        # First, figure out `self.id` if not provided.
        if not _id:
            if not hash_key:
                raise XRemoteError(
                    f"Tried to create DynKey with no id ({_id}) or no hash key ({hash_key})."
                )

            if require_full_key and need_range_key and not range_key:
                raise XRemoteError(
                    f"Tried to create DynKey with no id ({_id}) or no range key ({range_key})."
                )

            key_names = [(structure.dyn_hash_key_name, hash_key)]
            # Generate ID without delimiter to represent an entire hash-page (ie: any range value)
            if need_range_key and range_key is not None:
                key_names.append((range_name, range_key))

            keys = []
            for key_name, key_value in key_names:
                field = structure.get_field(key_name)
                converter = field.converter
                final_value = key_value
                if converter:
                    final_value = converter(
                        api,
                        Converter.Direction.to_json,
                        field,
                        key_value
                    )
                keys.append(final_value)

            _id = delimiter.join([str(x) for x in keys])
            object.__setattr__(self, 'id', _id)
        elif need_range_key and delimiter not in _id:
            raise XRemoteError(
                f"Tried to create DynKey with an `id` ({_id}) "
                f"that did not have delimiter ({delimiter}) in it. "
                f"This means we are missing the range-key part for field ({range_name}) "
                f"in the `id` that was provided.  Trying providing the id like this: "
                f'"{_id}{delimiter}'
                r'{range-key-value-goes-here}".'  # <-- want to directly-output the `{` part.
            )

        # If we got provided a hash-key directly, no need to continue any farther.
        if hash_key:
            if require_full_key and need_range_key and not range_key:
                raise XRemoteError(
                    f"Have hash_key ({hash_key}) without needed range_key while creating DynKey."
                )
            # We were provided the hash/range key already, as an optimization I don't use time
            # checking to see if they passed in the same values that they may have passed in `id`.
            return

        # They did not pass in hash_key, so we must parse the `id` they provided
        # and then set them on self.

        if not need_range_key:
            # If we don't need range key, there is no delimiter to look for.
            hash_key = _id
        else:
            split_id = _id.split(delimiter)
            if len(split_id) != 2:
                raise XRemoteError(
                    f"For dynamo table ({self.api.model_type}): Have id ({_id}) but delimiter "
                    f"({delimiter}) is either not present, or is in it more than once. "
                    f"'id' needs to contain exactly one hash and range key combined together "
                    f"with the delimiter, ie: 'hash-key-value{delimiter}range-key-value'. "
                    f"See xdynamo.dyn_connections documentation for more details on how "
                    f"this works."
                )
            # todo: Consider converting these `from_json`, like we convert `to_json`
            #   when we put the keys into the `id` (see above, where we generate `id` if needed)?
            hash_key = split_id[0]
            range_key = split_id[1]

        object.__setattr__(self, 'hash_key', hash_key)
        object.__setattr__(self, 'range_key', range_key)


class DynKeyType(Enum):
    """ Possible values for field option keys. """
    hash = EnumAuto()
    range = EnumAuto()


class _ProcessedQuery(Dict[str, Dict[str, Any]]):
    """
    A Dict/MutableMapping that will split the attribute name from the operator when given a
    `xurls.url.Query` via the `_ProcessedQuery.process_query` method.

    It will produce a dict/mapping that will map the attribute name in the query-key to
    another dict.  This second internal dict will map a normalized operator/condition name to
    it's value.

    Returning a `_ProcessedQuery` allows for an optimization.
    If we are asked to process a `_ProcessedQuery`, we will just return the result without
    having to actually check any of the values.

    Lets the RestClient pass around post-processed query results so it does not have to processes
    them a second time. This allows it to tell the difference between processed and unprocessed
    queries easily.

    We should only have keys with the attribute name as a string.
    Values should be a set with the operator(s) in them.
    """
    api: 'DynApi'

    @staticmethod
    def process_query(query: Query, *, api: 'DynApi') -> '_ProcessedQuery':
        if isinstance(query, _ProcessedQuery):
            # We already processed this in the past, just return it, no need to process it again.
            return query

        processed_query = _ProcessedQuery()
        processed_query.api = api
        for (k, v) in query.items():
            parts = k.split("__")
            name = parts[0]
            operator = None
            if len(parts) > 1:
                operator = parts[1]

            # When the operator is not provided, we guess the best one to use
            if operator is None:
                # If it's a list, we do the 'in' operator by default.
                if isinstance(v, list):
                    operator = "in"

            # Map alias operators to the standard one, otherwise keep current operator.
            operator = operator_alias_map.get(operator, operator)

            # Store value / operator in sub-dict...
            criterion = processed_query.setdefault(name, {})
            criterion[operator] = v
        return processed_query

    def generate_all_operator_values_for_name(self, name: str) -> Iterable[Tuple[str, str]]:
        """
        Produces a generator that will go though all operator/value combinations for
        the query param `name`. Each yield will be a tuple with operator as first,
        and value as second item in tuple.

        Args:
            name: name in query to use

        Returns: Tuple, first is operator and second is value.
            If the value is a list, will yield each value in list as a separate item.

        """
        for operator, values in self.get(name, ()).items():
            for value in xloop(values):
                yield operator, value

    def contains_only_keys(self):
        """ Return True if we have a key, and no other non-keys; Otherwise False. """
        api = self.api
        structure = api.structure
        key_names = {'id', structure.dyn_hash_key_name}
        range_key = structure.dyn_range_key_name
        if range_key:
            key_names.add(range_key)

        have_key = False
        for name in self:
            if name not in key_names:
                return False
            have_key = True

        return have_key

    _cached_dyn_keys = None

    def dyn_keys(self) -> Set[DynKey]:
        cached = self._cached_dyn_keys
        if cached is not None:
            return cached

        cached = self._generate_dyn_keys()
        self._cached_dyn_keys = cached
        return cached

    def _generate_dyn_keys(self) -> Set[DynKey]:
        """ Generate a set of DynKey's. This way they are uniquified. """
        api = self.api
        structure = api.structure
        hash_key = structure.dyn_hash_key_name
        range_key = structure.dyn_range_key_name
        dyn_keys = set()

        if hash_key and range_key in self and hash_key in self:
            hash_gen = self.generate_all_operator_values_for_name(hash_key)
            range_gen = list(self.generate_all_operator_values_for_name(range_key))

            # Go though every combination of hash + range keys....
            for hash_combo in hash_gen:
                for range_combo in range_gen:
                    range_operator = range_combo[0]
                    if range_operator == 'is_in':
                        range_operator = 'eq'

                    if range_operator == 'between':
                        next_range = next(range_gen, None)
                        if next_range[0] != 'between':
                            raise XRemoteError(
                                f"You must provide a second value for 'between' operator on range "
                                f"key ({range_key}), next value ({next_range[1]} had operator "
                                f"({next_range[0]})."
                            )
                        dyn_key = DynKey(
                            api=api,
                            hash_key=hash_combo[1],
                            range_key=[range_combo[1], next_range[1]],
                            range_operator='between'
                        )
                    else:
                        dyn_key = DynKey(
                            api=api,
                            hash_key=hash_combo[1],
                            range_key=range_combo[1],
                            range_operator=range_operator
                        )
                    dyn_keys.add(dyn_key)
        elif hash_key in self:
            for operator, value in self.generate_all_operator_values_for_name(hash_key):
                dyn_key = DynKey(
                    api=api,
                    hash_key=value,
                    # We want a key that represents an entire hash-page if there is a range-key.
                    require_full_key=False
                )
                dyn_keys.add(dyn_key)

        if 'id' != range_key and 'id' in self:
            for operator, value in self.generate_all_operator_values_for_name('id'):
                dyn_key = DynKey(
                    api=api,
                    id=value,
                    range_operator=operator
                )
                dyn_keys.add(dyn_key)

        return dyn_keys



import dataclasses
from typing import Optional, Type, Dict, TypeVar, Generic, Union
from xmodel import Field, Converter
from xdynamo.common_types import DynKeyType
from xsentinels.default import Default
import typing_inspect
from typing import Hashable
from xsentinels.sentinel import Sentinel

T = TypeVar('T')


class _HashSentinel(Sentinel):
    pass


class _RangeSentinel(Sentinel):
    pass


Hash = Union[T, None]
""" Used in type-hints to wrap other type-hints to declare that it could be that
    type or `Null` value. Indicates that a value of Null could be assigned.

    You can use it like so:

    >>> from null_type import Nullable, Null
    >>>
    >>> nullable_string: Nullable[str]
    >>>
    >>> # You can assign a `str` or a `Null` and it will be type-correct:
    >>> nullable_string = "hello!"
    >>> nullable_string = Null
"""

HashStr = Hash[str]


@dataclasses.dataclass
class _UnwrapResults:
    unwrapped_type: Type
    """ The type(s) with Hash/Range generic types filtered out.
    """
    is_hash: bool = False
    is_range: bool = False


def _unwrap_generic_hash_or_range_types(type_to_unwrap: Type, /) -> _UnwrapResults:
    """
    Returns the first non-Null or non-None type inside the optional/Union type as
    the `unwrapped_type` result.

    If the type passed in is not an optional/nullable/union type, then set returned
    `unwrapped_type` to the type unaltered.

    Args:
        type_to_unwrap: Type to inspect and unwrap the optionality/nullability/union from.
    Returns:
        UnwrapResults: With the unwrapped_type, and if type is Nullable and/or Optional.
            If the Union has more than one none-Null/None type in it, then we will return
            a Union with the None and Null types filtered out.
    """
    if not typing_inspect.is_generic_type(type_to_unwrap):
        return _UnwrapResults(type_to_unwrap)

    NoneType = type(None)
    saw_null = False
    saw_none = False
    types = []

    hint_union_sub_types = typing_inspect.get_args(type_to_unwrap)
    for sub_type in hint_union_sub_types:
        # if sub_type is NullType:
        #     saw_null = True
        #     continue

        if sub_type is NoneType:
            saw_none = True
            continue

        types.append(sub_type)

    if len(types) == 1:
        # No other non-Null/None types, use the type directly.
        unwrapped_type = types[0]
    else:
        # Construct final Union type with the None/Null filtered out.
        unwrapped_type = Union[tuple(types)]

    return UnwrapResults(unwrapped_type, is_nullable=saw_null, is_optional=saw_none)


class DynField(Field):
    dyn_key: Optional[DynKeyType] = Default

    def resolve_defaults(
            self,
            name,
            type_hint: Type,
            default_converter_map: Optional[Dict[Type, Converter]] = None,
            parent_field: "DynField" = None
    ):
        # pydoc3 will copy the parent-class doc-comment if left empty here;
        # that's exactly what I want so leaving doc-comment blank.
        super().resolve_defaults(
            name=name,
            type_hint=type_hint,
            default_converter_map=default_converter_map,
            parent_field=parent_field
        )

        if self.dyn_key:
            if not self.was_option_explicitly_set_by_user('include_in_repr'):
                self.include_in_repr = True


class HashField(DynField):
    dyn_key = DynKeyType.hash


class RangeField(DynField):
    dyn_key = DynKeyType.range
